# Copyright (c) 2011, Intel Corporation
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of Intel Corporation nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""bits module."""

import _bits
from _bits import *
import itertools
from collections import namedtuple
import string
import time

class cpuid_result(namedtuple('cpuid_result', ['eax', 'ebx', 'ecx', 'edx'])):
    __slots__ = ()

    def __repr__(self):
        return "cpuid_result(eax={eax:#08x}, ebx={ebx:#08x}, ecx={ecx:#08x}, edx={edx:#08x})".format(**self._asdict())

def cpuid(apicid, eax, ecx=0):
    """Run CPUID on the specified CPU. Return a namedtuple containing eax, ebx, ecx, and edx."""
    return cpuid_result(*_bits._cpuid(apicid, eax, ecx))

_grub_command_map = {}

def addr_alignment(addr):
    """Compute the maximum alignment of a specified address, up to 4."""
    return addr & 1 or addr & 2 or 4

PCI_ADDR_PORT = 0xCF8
PCI_DATA_PORT = 0xCFC

def _pci_op(bus, device, function, register, bytes):
    if bytes is None:
        bytes = addr_alignment(register)
    elif bytes not in [1,2,4]:
        raise ValueError("bytes must be 1, 2, or 4")
    outl(PCI_ADDR_PORT, ((1 << 31) | (bus << 16) | (device << 11) | (function << 8) | register) & ~3)
    return bytes, PCI_DATA_PORT + (register & 3)

def pci_read(bus, device, function, register, bytes=None):
    """Read a value of the specified size from the PCI device specified by bus:device.function register"""
    bytes, port = _pci_op(bus, device, function, register, bytes)
    return { 1: inb, 2: inw, 4: inl }[bytes](port)

def pci_write(bus, device, function, register, value, bytes=None):
    """Write a value of the specified size to the PCI device specified by bus:device.function register"""
    bytes, port = _pci_op(bus, device, function, register, bytes)
    { 1: outb, 2: outw, 4: outl }[bytes](port, value)

_pcie_base = None

def pcie_get_base():
    return _pcie_base

def pcie_set_base(memaddr):
    global _pcie_base
    if memaddr < 0 or memaddr >= 2**32:
        raise ValueError("PCIe base address out of supported range; must be below 4GB")
    _pcie_base = memaddr

def _pcie_op(bus, device, function, register, bytes, memaddr):
    if bytes is None:
        bytes = addr_alignment(register)
    elif bytes not in [1,2,4,8]:
        raise ValueError("bytes must be 1, 2, 4, or 8")
    if memaddr is None:
        memaddr = _pcie_base
        if memaddr is None:
            raise ValueError("PCIe base address not set")
    elif memaddr < 0 or memaddr >= 2**32:
        raise ValueError("PCIe base address out of supported range; must be below 4GB")
    return bytes, memaddr | (bus << 20) | (device << 15) | (function << 12) | register

def pcie_read(bus, device, function, register, bytes=None, memaddr=None):
    """Read a value of the specified size from the PCIe device specified by bus:device.function register"""
    bytes, addr = _pcie_op(bus, device, function, register, bytes, memaddr)
    return { 1: readb, 2: readw, 4: readl, 8: readq }[bytes](addr)

def pcie_write(bus, device, function, register, value, bytes=None, memaddr=None):
    """Write a value of the specified size to the PCIe device specified by bus:device.function register"""
    bytes, addr = _pcie_op(bus, device, function, register, bytes, memaddr)
    { 1: writeb, 2: writew, 4: writel, 8: writeq }[bytes](addr, value)

def register_grub_command(command, func, summary, description):
    """Register a new GRUB command, implemented using the given callable.
    The callable should accept a single argument, the list of argument strings.
    The arguments include the command name as [0]."""
    _bits._register_grub_command(command, summary, description)
    _grub_command_map[command] = func

def _grub_command_callback(args):
    try:
        return _grub_command_map[args[0]](args)
    except:
        import traceback
        traceback.print_exc()
        return False

_bits._set_grub_command_callback(_grub_command_callback)

_pyfs_files = {}

def _pyfs_dir(dirname):
    """_pyfs_dir(dirname) -> an iterable of (filename, is_directory) pairs,
    or None if not a directory"""
    if dirname != "/":
        return None
    return [(filename, 0) for filename in sorted(_pyfs_files.iterkeys())]

def _lookup(filename):
    components = filename.split("/")
    if len(components) != 2 or components[0] != '':
        return None
    return _pyfs_files.get(components[1])

def _pyfs_open(filename):
    """_pyfs_open(filename) -> the file size, or None if the file does not exist"""
    functions = _lookup(filename)
    if functions is None:
        return None
    do_open, do_read = functions
    return do_open()

def _pyfs_read(filename, offset, size):
    """_pyfs_read(filename, offset, size) -> size bytes starting at offset,
    as a string"""
    functions = _lookup(filename)
    if functions is None:
        return None
    do_open, do_read = functions
    return do_read(offset, size)

_bits._set_pyfs_callbacks(_pyfs_dir, _pyfs_open, _pyfs_read)

def pyfs_add(filename, do_open, do_read):
    """Add a file to the (python) filesystem.

    Attempts to open and read the file will get passed to the specified open
    and read functions, which should have the following signatures:

    do_open(): return the length of the file.

    do_read(offset, size): return the data from offset to offset+size as a
    string."""
    assert "/" not in filename
    assert "\0" not in filename
    assert _pyfs_files.get(filename) is None
    assert callable(do_open)
    assert callable(do_read)
    _pyfs_files[filename] = do_open, do_read

def bsp_apicid():
    """Returns the BSP's APIC ID."""
    cpulist = cpus()
    return cpulist[0]

def socket_index(apic_id):
    """Returns the socket portion of the APIC ID"""
    if apic_id is None:
        return None
    if cpuid(apic_id,0x0)[0] < 0xb:
        return None
    eax, ebx, ecx, edx = cpuid(apic_id,0xb,1)
    x2apic_id = edx
    socket_index = x2apic_id >> (eax & 0x1f);
    return socket_index

def socket_apic_ids():
    """Returns a mapping to unique socket index with the list of APIC IDs"""
    uniques = {}
    cpulist = cpus()
    for apicid in sorted(cpulist):
        uniques.setdefault(socket_index(apicid), []).append(apicid)
    return uniques

def apicid_to_index():
    """Returns a reverse mapping from APIC ID to CPU number"""
    cpulist = cpus()
    return dict([(apicid, i) for (i, apicid) in enumerate(cpulist)])

def cpu_frequency():
    IA32_MPERF_MSR = 0xE7
    IA32_APERF_MSR = 0xE8

    if cpuid(bsp_apicid(), 0).eax < 6:
        # CPUID Leaf 6 is not supported
        return None

    if cpuid(bsp_apicid(), 6).ecx & 1 == 0:
        # MPERF/APERF MSRs are not supported
        return None

    cpulist = cpus()
    for apicid in cpulist:
        set_mwait(apicid, True, 0x20)

    if wrmsr(bsp_apicid(), IA32_MPERF_MSR, 0) is None:
        # Writing of IA32_MPERF MSR caused a GPF
        return None

    if wrmsr(bsp_apicid(), IA32_APERF_MSR, 0) is None:
        # Writing of IA32_APERF MSR caused a GPF
        return None

    # Needs to busywait, not sleep
    start = time.time()
    while (time.time() - start < 1):
        pass

    mperf_hz = rdmsr(bsp_apicid(), IA32_MPERF_MSR)
    aperf_hz = rdmsr(bsp_apicid(), IA32_APERF_MSR)

    return mperf_hz, aperf_hz

def print_hz(hz):
    temp = hz / (1000.0 * 1000 * 1000)
    if abs(temp) >= 1:
        return '{:.03f} GHz'.format(temp)
    temp = hz / (1000.0 * 1000)
    if abs(temp) >= 1:
        return '{:.03f} MHz'.format(temp)
    temp = hz / 1000.0
    if abs(temp) >= 1:
        return '{:.03f} kHz'.format(temp)
    return '{} Hz'.format(hz)

def print_cpu_freq():
    mperf_hz, aperf_hz = cpu_frequency()
    delta_hz = aperf_hz - mperf_hz

    print "Frequency = {} (MPERF)  {} (APERF)  {} (delta)".format(print_hz(mperf_hz), print_hz(aperf_hz), print_hz(delta_hz))

def grouper(n, iterable, fillvalue=None):
    "grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return itertools.izip_longest(fillvalue=fillvalue, *args)

def dumpmem(mem, addr=0):
    """Dump hexadecimal and printable ASCII bytes for a memory buffer"""
    for offset, chunk in zip( range(0, len(mem), 16), grouper(16, mem) ):
        print "{:08x}: ".format(addr + offset),
        print " ".join("  " if x is None else "{:02x}".format(ord(x)) for x in chunk),
        s = "  "
        for x in chunk:
            if x is None:
                s += ' '
            elif x in string.letters or x in string.digits or x in string.punctuation:
                s += x
            else:
                s += '.'
        print s
